-- Initialize Per Character values
chatEnabled = false
floatingEnabled = true
averageEnabled = {true, true, true}
killAverage = nil 
questAverage = nil

-- Session values
local currentLevel, xpTotal, xpCurrent, xpRemaining, xpRested
local lastMobName, lastMobXP, lastQuestXP
local killsRemaining, questsRemaining
local mobAverageTable = {}
local questAverageTable = {}

-- Waterfall GUI Config
local waterfall = AceLibrary("Waterfall-1.0")

----------------------
-- Initialization
----------------------
-- Main entry point
function MainOnload()
	-- Disable if the character is at max level
	currentLevel = UnitLevel("player")
	if(currentLevel == 80) then
		return;
	end
	
	-- Check if the client language is suported
	if(GetLocale() ~= "enUS") then
		DEFAULT_CHAT_FRAME:AddMessage("XToLevel could not be loaded. Client language is not supported.");
		return;
	end
	
	-- Register Events
	this:RegisterEvent("CHAT_MSG_COMBAT_XP_GAIN");
	this:RegisterEvent("PLAYER_LOGIN");
	this:RegisterEvent("PLAYER_LEVEL_UP");
	
	-- Initialize GUI config
	ConfigWindow_Init();
	
	-- Register slash commands
	SLASH_XTOLEVEL1 = "/xtolevel";
	SLASH_XTOLEVEL2 = "/xtl";
	SlashCmdList["XTOLEVEL"] = SlashHandler;
end
function SlashHandler()
	waterfall:Open('ConfigWindow', 'messages')
end
function ConfigWindow_Init()
	waterfall:Register('ConfigWindow',
		'title','XToLevel Config',
		'colorR', 0.0,
		'colorG', 0.5,
		'colroB', 0.25,
		'tree', 
		{
			{
				text = 'Messages',
				id = 'messages',
				isOpen = true,
				isSelected = true
			},
			{
				text = 'Average display',
				id = 'average',
				isOpen = true,
				isSelected = true
			}
		},
	 'children', 
		function(id)
			if id=='messages' then
				waterfall:AddControl(
					'type', 'checkbox',
					'text', "Show in default chat window",
					'checked', chatEnabled,
					'fullRefresh', true,
					'setFunc', function() chatEnabled = not chatEnabled end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', "Show floating message",
					'checked', floatingEnabled,
					'fullRefresh', true,
					'setFunc', function() floatingEnabled = not floatingEnabled end
		        )
				return 'Messages'
			elseif id=='average' then
				waterfall:AddControl(
					'type', 'checkbox',
					'text', "Enable average display",
					'checked', averageEnabled[1],
					'fullRefresh', true,
					'setFunc', function() ToogleDisplay() end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', "Show mob average",
					'checked', averageEnabled[2],
					'fullRefresh', true,
					'setFunc', function() ToogleMobDisplay() end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', "Show quest average",
					'checked', averageEnabled[3],
					'fullRefresh', true,
					'setFunc', function() ToogleQuestDisplay() end
		        )
				return 'Avarage display'
			end
		end 
	)
end

----------------------
-- GUI Config callbacks
----------------------
function ToogleDisplay()
	averageEnabled[1] = not averageEnabled[1];
	if averageEnabled[0] then
		XToLevelDisplay:Show();
	else
		XToLevelDisplay:Hide();
	end
end
function ToogleMobDisplay()
	averageEnabled[2] = not averageEnabled[2];
end
function ToogleQuestDisplay()
	averageEnabled[3] = not averageEnabled[3];
end

----------------------
-- Display functions
----------------------
function DisplaySet(mobs, quests)
	XToLevelDisplayLabelText:SetText("XToLevel:");
	
	if mobs ~= nil and averageEnabled[2] then
		XToLevelDisplayKillText:SetText(" Mobs: ".. mobs);
	end
	if quests ~= nil and averageEnabled[3] then
		if (mobs == nil or averageEnabled[2] == false) then -- Use mob box if no mob text was passed
			XToLevelDisplayKillText:SetText(" Quests: ".. quests);
		else
			XToLevelDisplayQuestText:SetText(" Quests: ".. quests);
		end
	end
end

----------------------
-- Misc functions
----------------------
-- Used to find the Zero Difference used to calculate XP from low level mobs
function GetZeroDifference(charLevel)
	for i,v in ipairs(ZD_Table) do
		if charLevel >= v[1] and charLevel <= v[2] then
			return v[3];
		end
	end
end

-- Returns a boolean value, indicating whether your character is in the outlands
function GetZoneID()
	local currentZone = GetRealZoneText();
	local continentNames, key, val = { GetMapContinents() } ;
	for key, val in pairs(continentNames) do
		local continentZones = { GetMapZones(key) };
		for i,zone in ipairs(continentZones) do
			if zone == currentZone then
				return i
			end
		end
	end
	return nil
end

-- Extracts the mob name and XP gained from the combat XP message
function ExtractKillInfoReg(message)
	local pattern, mob, xp;
	--- Check if this is a quest or a mob
	if(string.find(message, ",") == nil) then
		pattern = "You gain (%d+) experience";
		if(GetLocale() == "deDE") then
			pattern = "Ihr bekommt (%d+) Erfahrung";
		elseif (GetLocale() == "frFR") then
			pattern = "Vous gagnez (%d+) points d'experience";
		end
		xp = strmatch(message, pattern);
	else
		pattern = "(.+) dies, you gain (%d+) experience";
		if(GetLocale() == "deDE") then
			pattern = "(.+) stirbt, Ihr bekommt (%d+) Erfahrung";
		elseif (GetLocale() == "frFR") then
			pattern = "(.+) succombe, vous gagnez (%d+) points d'experience";
		end
		mob, xp = strmatch(message, pattern);
	end
	
	return mob, xp;
end
----------------------
-- Experience calculations
----------------------
-- Get total rested bonus available for this level
function GetRestedToLevel()
	local xpToLevel = xpTotal - xpCurrent;
	local restedToLevel = 0;
	if(xpRested == 0) then
		restedToLevel = 0;
	elseif(xpToLevel / 2 < xpRested) then
		restedToLevel = xpToLevel / 2;
	else
		restedToLevel = xpRested;
	end
	return restedToLevel
end
-- Gets the total XP for the given level
function GetXPMax(level)
	local unit = UnitXPMax("player")
	if(level < 70 and unit < XP_TABLE[level][3]) then
		return XP_TABLE[level][3]
	else
		return unit
	end
end
-- Calculates the XP gained from killing a mob
function GetMobXP(charLevel, mobLevel)
	-- Find the gray level
	local grayLevel = 0
	if charLevel < 6 then
		grayLevel = 0
	elseif charLevel < 40 then
		grayLevel = charLevel - floor(charLevel / 10) - 5
	elseif charLevel < 60 then
		grayLevel = charLevel - floor(charLevel / 5) - 1
	else
		grayLevel = charLevel - 9
	end
	
	-- Check if mob is gray
	if mobLevel <= grayLevel then
		return 0
	end
	
	-- Set the add value (higher for outlands)
	local addValue = 45
	if GetZoneID() == 3 then
		addValue = 235
	end
	
	-- Calculate XP
	local XP = 0
	if(charLevel == mobLevel) then
		XP = (charLevel * 5) + addValue
	elseif (charLevel < mobLevel) then
		-- Find max mob level. All mobs higher than (charLevel + 4) are worth the same amount.
		local maxMobLevel = mobLevel
		if mobLevel > charLevel + 4 then 
			maxMobLevel = charLevel + 4 
		end
		XP = ((charLevel * 5 + addValue)) * (1 + (0.05 * (mobLevel - charLevel)))
	elseif (charLevel > mobLevel) then
		local ZD = GetZeroDifference(charLevel)
		XP = ((charLevel * 5 + addValue)) * (1 - ((charLevel - mobLevel) / ZD))
	end
	
	return XP
end
-- Gets the Kills and Quests needed to reach the next level
function GetXToLevel()
	local killsToLevel, questsToLevel

	-- Calculate Mob kills
	local xpPerMob = GetMobAverage()
	if xpPerMob == nil then
		if killAverage == nil then
			xpPerMob = GetMobXP(currentLevel, currentLevel)
		else
			xpPerMob = killAverage
		end
	end
	
	local restedToLevel = GetRestedToLevel()
	killsToLevel = ceil(((xpTotal - xpCurrent) - restedToLevel) / xpPerMob)
	
	-- Calculate Quest handins
	local xpPerQuest = GetQuestAverage()
	if xpPerQuest == nil and questAverage == nil then
		questsToLevel = nil
	elseif questAverage ~= nil then
		questsToLevel = ceil((xpTotal - xpCurrent) / questAverage)
	else
		questsToLevel = ceil((xpTotal - xpCurrent) / xpPerQuest)
	end
	
	return killsToLevel, questsToLevel
end

----------------------
-- Average calculations
----------------------
function GetMobAverage()
	local tableSize = # mobAverageTable
	if tableSize > 0 then
		local total = 0
		local count = 0
		for i,v in ipairs(mobAverageTable) do
			total = total + v
			count = count + 1
		end
		return floor(total / count)
	else
		return nil
	end
end
function AddToMobAverage(xpGained)

	local bufferTable = {}
	bufferTable[1] = xpGained
	local maxStored = 10
	for i,v in ipairs(mobAverageTable) do
		if i >= maxStored then break end
		bufferTable[i+1] = v
	end
	mobAverageTable = bufferTable
end
function ResetMobAverage()
	if # mobAverageTable > 0 then
		mobAverageTable = { GetMobAverage() }
	else
		mobAverageTable = { }
	end
end
function GetQuestAverage()
	if # questAverageTable > 0 then
		local total = 0
		local count = 0
		for i,v in ipairs(questAverageTable) do
			total = total + v
			count = count + 1
		end
		return floor(total / count)
	else
		return nil
	end
end
function AddToQuestAverage(xpGained)
	local bufferTable = {}
	bufferTable[1] = xpGained
	local maxStored = 10
	for i,v in ipairs(questAverageTable) do
		if i >= maxStored then break end
		bufferTable[i+1] = v
	end
	questAverageTable = bufferTable
end
function ResetQuestAverage()
	if # questAverageTable > 0 then
		questAverageTable = { GetQuestAverage() }
	else
		questAverageTable = { }
	end
end

----------------------
-- Event callbacks
----------------------
function MainOnPlayerLogin()
	-- Get XP values
	xpTotal = GetXPMax(UnitLevel("player"))
	xpCurrent = UnitXP("player") or 0
	xpRested = GetXPExhaustion() or 0
	xpRested = (xpRested / 2) -- Divide so it only reflects the additional XP
	
	-- Get X to level
	killsRemaining, questsRemaining = GetXToLevel()
	
	-- Show display
	if averageEnabled[1] then
		XToLevelDisplay:Show();
		DisplaySet(killsRemaining, questsRemaining);
	else
		XToLevelDisplay:Hide();
	end
end
function OnPlayerLevelUp(newLevel)
	-- Set new level and XP values
	currentLevel = newLevel
	xpTotal = GetXPMax(newLevel)
	xpCurrent = 0 -- Set to 0 to prevent old data from being used
	xpRested = GetXPExhaustion() or 0
	xpRested = (xpRested / 2) -- Divide so this only reflects the additional XP
	
	-- Reset averages
	ResetMobAverage()
	ResetQuestAverage()
	
	-- Get kills and quests to level
	killsRemaining, questsRemaining = GetXToLevel()
	
	-- Set display
	DisplaySet(killsRemaining, questsRemaining)
end
function OnChatXPGain(msg)
	-- Update the XP info
	xpTotal = UnitXPMax("player")
	xpCurrent = UnitXP("player")
	xpRested = GetXPExhaustion() or 0
	xpRested = (xpRested / 2) -- Divide so this only reflects the additional XP

	-- Extract the info from the message
	local mobName, xpGained = ExtractKillInfoReg(arg1);
	xpGained = tonumber(xpGained)
	
	-- Add the new XP to the currend XP
	xpCurrent = xpCurrent + xpGained
	
	-- Level reached
	local msg = "n/a"
	if(xpCurrent >= xpTotal) then
		msg = "Level reached. Congratulations"
	-- Quest turn-in
	elseif(mobName == nil) then
		-- Calculate the ammount of quest turnins to level
		--  (TotalLevelXP - XPSoFar - GainedXP) / GainedXP
		local quests = ceil((xpTotal - xpCurrent) / xpGained)
		
		-- Create message
		msg = quests .. " more quests like that needed to level"
		
		--  Add to average
		AddToQuestAverage(xpGained)
		questAverage = GetQuestAverage()
	
	-- Mob kill
	else
		-- Check if the XP gained was rested and remove the rested bonus if needed
		if(xpRested > 0) then
			if(xpRested * 2 < xpGained) then
				xpGained = xpGained - xpRested
			else
				xpGained = xpGained / 2
			end
		end
		
		-- Get How much rested till next level
		local restedToLevel = GetRestedToLevel()
		
		-- Calculate the ammount of mobs needed
		--  (TotalLevelXP - XPSoFar - (RestedBonus)) / GainedXP
		local kills = ceil(((xpTotal - xpCurrent) - restedToLevel) / xpGained)
		
		-- Create Message
		msg = kills .." ".. mobName .." needed to level"
		
		-- Add to average
		AddToMobAverage(xpGained)
		killAverage = GetMobAverage()
	end
		
	-- Send Messages
	if chatEnabled then
		DEFAULT_CHAT_FRAME:AddMessage(msg, 0, 100, 200, 55, 4);
	end
	if floatingEnabled then
		UIErrorsFrame:AddMessage(msg, 0, 100, 200, 55, 4);
	end
	
	-- Update average display and per character values
	killsRemaining, questsRemaining = GetXToLevel()
	DisplaySet(killsRemaining, questsRemaining)
end

----------------------
-- Event handler
----------------------
function MainOnEvent()
	if (event == "PLAYER_LOGIN") then
		MainOnPlayerLogin()
	elseif (event == "PLAYER_LEVEL_UP") then
		OnPlayerLevelUp(arg1)
	elseif (event == "CHAT_MSG_COMBAT_XP_GAIN") then
		OnChatXPGain(arg1)
	end
end
